//===- reifier/ReifyingVisitor.h - Reifies interesting program characteristics -*- C++ -*-===//
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Aspicere2.
 *
 * The Initial Developer of the Original Code is
 * the Ghislain Hoffman Software Engineering Lab, INTEC, University Ghent.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Bram Adams <bram_DOT_adams_AT_ugent_DOT_be>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef REIFYINGVISITOR_H
#define REIFYINGVISITOR_H

#include "prolog/QueryEngine.h"
#include "llvm/Support/InstVisitor.h"

class ReifyingVisitor: public llvm::InstVisitor<ReifyingVisitor> {
 public:
  ReifyingVisitor(QueryEngine* eng):engine(eng){}

  virtual ~ReifyingVisitor(){}

  /**
   * Only declaration or definition but not both in LLVM IR?
   *
   * methodDef(id,name) and/or methodDecl(id,name)
   * executionShadow(id)%id is same as definition id for now
   * methodType(id,type)
   * methodArg(id,methoddef/decl_id,type,pos,nextpos,name)
   */
  void visitFunction(llvm::Function &F);//execution

  /**
   * varDecl(id,type,name)
   */
  void visitModule(llvm::Module &M);//function declarations and global vars

  /**
   * callShadow(id,called_methoddef/decl_id,enclosing_shadow_id)
   * callActual(id,call_id,pos,nextpos)
   */
  void visitCallInst(llvm::CallInst &I);//call  

  /**
   * callShadow(id,called_methoddef/decl_id,enclosing_shadow_id)
   * callActual(id,call_id,pos,nextpos)
   */
  void visitInvokeInst(llvm::InvokeInst &I);//call

  /**
   * callShadow(id,called_methoddef/decl_id,enclosing_shadow_id)
   * callActual(id,call_id,pos,nextpos)
   */
  void visitVAArgInst(llvm::VAArgInst &I);
 private:
  QueryEngine* engine;
};

#endif
